;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "auto-dark" "0.13.10"
  "Automatically set the dark-mode theme based on system status."
  '((emacs "24.4"))
  :url "https://github.com/LionyxML/auto-dark-emacs"
  :commit "6d1e8d2fc493dccbf05c9191611805c7e7881c70"
  :revdesc "6d1e8d2fc493"
  :keywords '("macos" "windows" "linux" "themes" "tools" "faces")
  :authors '(("Tim Harper" . "timcharperatgmaildotcom")
             ("Vincent Zhang" . "seagle0128@gmail.com")
             ("Jonathan Arnett" . "jonathan.arnett@protonmail.com")
             ("Greg Pfeil" . "greg@technomadic.org"))
  :maintainers '(("Tim Harper" . "timcharperatgmaildotcom")
                 ("Vincent Zhang" . "seagle0128@gmail.com")
                 ("Jonathan Arnett" . "jonathan.arnett@protonmail.com")
                 ("Greg Pfeil" . "greg@technomadic.org")))
